/*
 * Copyright (C) 1993-2000 by Darren Reed.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and due credit is given
 * to the original author and the contributors.
 */

/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com for
 * details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unix.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <sys/time.h>
#include <net/if.h>
#include <netinet/ip.h>
#include <netdb.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include "netinet/ip_compat.h"
#include "netinet/ip_fil.h"
#include "netinet/ip_nat.h"
#include "netinet/ip_state.h"
#include "netinet/ipf.h"
#include "netinet/ipl.h"

#if !defined(lint) && defined(RCCS)
static const char sccsid[] = "@(#)ipf.c	1.23 6/5/96 (C) 1993-2000 Darren Reed";
static const char rcsid[] = "@(#)$Id: ipf.c,v 2.10 2000/03/13 22:10:23 darrenr Exp $";
#endif

extern void    ipf_mainloop __P((int));
extern ipnat_t *natparse    __P((char *, int));
extern void    printnat     __P((ipnat_t *, int, void *));
extern int     iplattach    __P((void));
extern void    procfile __P((char *, char *));

extern char    *optarg;

int	main __P((int, char *[]));

int	opts = 0;
#ifdef	USE_INET6
int	use_inet6 = 0;
#endif


static	char	*getline __P((char *, size_t, FILE *));
static	void	usage __P((void));
static	void	showversion __P((void));

static void usage()
{
	fprintf(stderr, "usage: ipfd [-dVv] [-l block|pass|nomatch] [-f i|o|a|s|S] [-f conf_file]");
	exit(1);
}

#if QNX < 600
int main(argc,argv)
int argc;
char *argv[];
{
	int c;

	if ((c = iplattach()) != EOK)
	  return (c== -1 ? errno : c);

	while ((c = getopt(argc, argv, "df:lvV")) != -1) {
		switch (c)
		{
		  case '?' :
			usage();
		  case 'd' :
			opts |= OPT_DEBUG;
			break;
		  case 'f' :
			procfile(argv[0], optarg);
			break;
		  case 'v' :
			opts |= OPT_VERBOSE;
			break;
		  case 'V' :
			showversion();
			break;
		}
	}

	ipf_mainloop(opts);

	return (0);
	/* NOTREACHED */
}
#endif
void	procfile(name, file)
char	*name, *file;
{
	FILE	*fp;
	char	line[513], *s, natline[513];
	struct	frentry	*fr;
	u_int	add;
	int     linenum = 0;
	ipnat_t *np;

//	initparse();

	if (!strcmp(file, "-"))
		fp = stdin;
	else if (!(fp = fopen(file, "r"))) {
		fprintf(stderr, "%s: fopen(%s) failed: %s\n", name, file,
			STRERROR(errno));
		exit(1);
	}

	while (getline(line, sizeof(line), fp)) {
	        linenum++;
		/*
		 * treat CR as EOL.  LF is converted to NUL by getline().
		 */
		if ((s = index(line, '\r')))
			*s = '\0';
		/*
		 * # is comment marker, everything after is a ignored
		 */
		if ((s = index(line, '#')))
			*s = '\0';

		if (!*line)
			continue;

		if (opts & OPT_VERBOSE)
			(void)fprintf(stderr, "[%s]\n", line);

		memcpy(natline, line, strlen(line));
		fr = parse(line, linenum);
		(void)fflush(stdout);

		if (fr) {
			add = (u_int)fr->fr_hits ? SIOCINAFR : SIOCADAFR;
			if (fr->fr_hits)
				fr->fr_hits--;
			if (fr && (opts & OPT_VERBOSE))
				printfr(fr);

			if (opts & OPT_DEBUG)
				binprint(fr);

			if (ipf_ioctl((caddr_t)&fr, add, FWRITE, 0) == -1)
			  perror("ioctl(add/insert rule)");
		} else if ((np = natparse(natline, linenum)) != NULL) {
			if ((opts & OPT_VERBOSE) && np)
			  printnat(np, opts & OPT_VERBOSE, NULL);
			if (nat_ioctl((caddr_t)&np, SIOCADNAT, FWRITE, 0) == -1)
			  perror("ioctl(SIOCADNAT)");
		} else {
			fprintf(stderr, "%d: syntax error in \"%s\"\n", linenum, line);
		}
	}
	if (ferror(fp) || !feof(fp)) {
		fprintf(stderr, "%s: %s: file error or line too long\n",
		    name, file);
		exit(1);
	}
	(void)fclose(fp);
}

/*
 * Similar to fgets(3) but can handle '\\' and NL is converted to NUL.
 * Returns NULL if error occured, EOF encounterd or input line is too long.
 */
static char *getline(str, size, file)
register char	*str;
size_t	size;
FILE	*file;
{
	char *p;
	int s, len;

	do {
		for (p = str, s = size;; p += len, s -= len) {
			/*
			 * if an error occured, EOF was encounterd, or there
			 * was no room to put NUL, return NULL.
			 */
			if (fgets(p, s, file) == NULL)
				return (NULL);
			len = strlen(p);
			p[len - 1] = '\0';
			if (p[len - 1] != '\\')
				break;
			size -= len;
		}
	} while (*str == '\0' || *str == '\n');
	return (str);
}

static void showversion()
{
	printf("ipfd: %s \n", IPL_VERSION);
}
