/*
 * Copyright (C) 1997-2000 by Darren Reed.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and due credit is given
 * to the original author and the contributors.
 *
 * $Id: ip_log.c,v 2.5 2000/03/13 22:10:21 darrenr Exp $
 */
#ifdef	IPFILTER_LOG
# include <sys/param.h>
# include <errno.h>
# include <sys/types.h>
# include <sys/file.h>
# include <fcntl.h>
# include <sys/ioctl.h>
# include <sys/time.h>
# include <sys/uio.h>
# include <sys/dir.h>
# include <sys/cdefs.h>
# include <sys/mbuf.h>
# include <sys/protosw.h>
# include <sys/socket.h>
# include <sys/kernel.h>

# include <net/if.h>
# include <net/route.h>
# include <netinet/in.h>
# include <netinet/in_var.h>
# include <netinet/in_systm.h>
# include <netinet/ip.h>
# include <netinet/tcp.h>
# include <netinet/udp.h>
# include <netinet/ip_icmp.h>
# include <netinet/ip_var.h>
# include <syslog.h>
# include "netinet/ip_compat.h"
# include <netinet/tcpip.h>
# include "netinet/ip_fil.h"
# include "netinet/ip_proxy.h"
# include "netinet/ip_nat.h"
# include "netinet/ip_frag.h"
# include "netinet/ip_state.h"
# include "netinet/ip_auth.h"

# include "ipf_res.h"

# ifndef MIN
#  define	MIN(a,b)	(((a)<(b))?(a):(b))
# endif


iplog_t	**iplh[IPL_LOGMAX+1], *iplt[IPL_LOGMAX+1], *ipll[IPL_LOGMAX+1];
size_t	iplused[IPL_LOGMAX+1];
static fr_info_t	iplcrc[IPL_LOGMAX+1];


/*
 * Initialise log buffers & pointers.  Also iniialised the CRC to a local
 * secret for use in calculating the "last log checksum".
 */
void ipflog_init()
{
	int	i;

	for (i = IPL_LOGMAX; i >= 0; i--) {
		iplt[i] = NULL;
		ipll[i] = NULL;
		iplh[i] = &iplt[i];
		iplused[i] = 0;
		bzero((char *)&iplcrc[i], sizeof(iplcrc[i]));
	}
}


/*
 * ipflog
 * Create a log record for a packet given that it has been triggered by a
 * rule (or the default setting).  Calculate the transport protocol header
 * size using predetermined size of a couple of popular protocols and thus
 * how much data to copy into the log, including part of the data body if
 * requested.
 */
int ipflog(flags, ip, fin, m)
u_int flags;
ip_t *ip;
fr_info_t *fin;
mb_t *m;
{
	ipflog_t ipfl;
	register size_t mlen, hlen;
	size_t sizes[2];
	void *ptrs[2];
	int types[2];
	u_char p;
	struct ifnet *ifp = fin->fin_ifp;

	/*
	 * calculate header size.
	 */
	hlen = fin->fin_hlen;
	if (fin->fin_off == 0) {
		p = fin->fin_fi.fi_p;
		if (p == IPPROTO_TCP)
			hlen += MIN(sizeof(tcphdr_t), fin->fin_dlen);
		else if (p == IPPROTO_UDP)
			hlen += MIN(sizeof(udphdr_t), fin->fin_dlen);
		else if (p == IPPROTO_ICMP) {
			struct icmp *icmp;

			icmp = (struct icmp *)fin->fin_dp;
	 
			/*
			 * For ICMP, if the packet is an error packet, also
			 * include the information about the packet which
			 * caused the error.
			 */
			switch (icmp->icmp_type)
			{
			case ICMP_UNREACH :
			case ICMP_SOURCEQUENCH :
			case ICMP_REDIRECT :
			case ICMP_TIMXCEED :
			case ICMP_PARAMPROB :
				hlen += MIN(sizeof(struct icmp) + 8,
					    fin->fin_dlen);
				break;
			default :
				hlen += MIN(sizeof(struct icmp),
					    fin->fin_dlen);
				break;
			}
		}
	}
	/*
	 * Get the interface number and name to which this packet is
	 * currently associated.
	 */
//	strncpy(ipfl.fl_ifname, ifp->if_xname, IFNAMSIZ);
	strncpy(ipfl.fl_ifname, (char *)ifp, IFNAMSIZ);
	mlen = (flags & FR_LOGBODY) ? MIN(fin->fin_plen - hlen, 128) : 0;
	ipfl.fl_plen = (u_char)mlen;
	ipfl.fl_hlen = (u_char)hlen;
	ipfl.fl_rule = fin->fin_rule;
	ipfl.fl_group = fin->fin_group;
	if (fin->fin_fr != NULL)
		ipfl.fl_loglevel = fin->fin_fr->fr_loglevel;
	else
		ipfl.fl_loglevel = 0xffff;
	ipfl.fl_flags = flags;
	ptrs[0] = (void *)&ipfl;
	sizes[0] = sizeof(ipfl);
	types[0] = 0;
	ptrs[1] = m;
	sizes[1] = hlen + mlen;
	types[1] = 1;
	return ipllog(IPL_LOGIPF, fin, ptrs, sizes, types, 2);
}


/*
 * ipllog
 */
int ipllog(dev, fin, items, itemsz, types, cnt)
int dev;
fr_info_t *fin;
void **items;
size_t *itemsz;
int *types, cnt;
{
	caddr_t buf, s;
	iplog_t *ipl;
	size_t len;
	int i;
	struct reader_list_t *rlist;
	struct reader        *r;
 
	/*
	 * Check to see if this log record has a CRC which matches the last
	 * record logged.  If it does, just up the count on the previous one
	 * rather than create a new one.
	 */
	MUTEX_ENTER(&ipl_mutex);
	if (fin != NULL) {
		if ((ipll[dev] != NULL) &&
		    bcmp((char *)fin, (char *)&iplcrc[dev], FI_CSIZE) == 0) {
			ipll[dev]->ipl_count++;
			MUTEX_EXIT(&ipl_mutex);
			return 1;
		}
		bcopy((char *)fin, (char *)&iplcrc[dev], FI_CSIZE);
	} else
		bzero((char *)&iplcrc[dev], FI_CSIZE);
	MUTEX_EXIT(&ipl_mutex);

	/*
	 * Get the total amount of data to be logged.
	 */
	for (i = 0, len = sizeof(iplog_t); i < cnt; i++)
		len += itemsz[i];

	/*
	 * check that we have space to record this information and can
	 * allocate that much.
	 */
	KMALLOCS(buf, caddr_t, len);
	if (!buf)
		return 0;
	MUTEX_ENTER(&ipl_mutex);
	if ((iplused[dev] + len) > IPLLOGSIZE) {
		MUTEX_EXIT(&ipl_mutex);
		KFREES(buf, len);
		return 0;
	}
	iplused[dev] += len;
	ipl_stat[dev].st_size += len;
	MUTEX_EXIT(&ipl_mutex);

	/*
	 * advance the log pointer to the next empty record and deduct the
	 * amount of space we're going to use.
	 */
	ipl = (iplog_t *)buf;
	ipl->ipl_magic = IPL_MAGIC;
	ipl->ipl_count = 1;
	ipl->ipl_next = NULL;
	ipl->ipl_dsize = len;
	ipl->ipl_sec = time(NULL);
//	microtime((struct timeval *)&ipl->ipl_sec);

	/*
	 * Loop through all the items to be logged, copying each one to the
	 * buffer.  Use bcopy for normal data or the mb_t copyout routine.
	 */
	for (i = 0, s = buf + sizeof(*ipl); i < cnt; i++) {
		if (types[i] == 0)
			bcopy(items[i], s, itemsz[i]);
		else if (types[i] == 1) {
			m_copydata(items[i], 0, itemsz[i], s);
		}
		s += itemsz[i];
	}
	MUTEX_ENTER(&ipl_mutex);
	ipll[dev] = ipl;
	*iplh[dev] = ipl;
	iplh[dev] = &ipl->ipl_next;
	MUTEX_EXIT(&ipl_mutex);
//	wakeup(&iplh[dev]);

	/* wakeup reader if possiable */
	rlist = &reader_list[dev];
	while (rlist->head != rlist->tail)
	{
		r = &rlist->readers[rlist->head];

		if (r->pid) {
			len = ipflog_read(r->pid, r->ocb->unit, &r->msg);
			switch (len) {
			  case 0:
			  case -1:
				/* to be block, or error in read, left the
				 * reader on list (head did not move)
				 */
				return 1;
			  default:
				((struct _io_read_reply *)&r->msg)->status = EOK;
				((struct _io_read_reply *)&r->msg)->nbytes = len;
				r->ocb->rd = NULL;
				Reply(r->pid, &r->msg, sizeof(struct _io_read_reply));
				break;
			}
		}

		if (++rlist->head == rlist->maxlen)
		  rlist->head = 0;
	}
	return 1;
}


int ipflog_read(pid, unit, msg)
pid_t pid;
int   unit;
struct _io_read *msg;
{
	size_t dlen, copied;
	int error = 0;
	iplog_t *ipl;

	/*
	 * Sanity checks.  Make sure the minor # is valid and we're copying
	 * a valid chunk of data.
	 */
	if (IPL_LOGMAX < unit)
		return ENXIO;
	if (!msg->nbytes)
		return 0;
	if ((msg->nbytes < sizeof(iplog_t)) ||
	    (msg->nbytes > IPLLOGSIZE))
		return EINVAL;
 	
	/*
	 * if no log data, block the reader
	 */
	if (!iplused[unit] || !iplt[unit])
	  return 0;

	for (copied = 0; (ipl = iplt[unit]); copied += dlen) {
		dlen = ipl->ipl_dsize;
		if (copied > msg->nbytes)
			break;
		/*
		 * Don't hold the mutex over the uiomove call.
		 */
		iplt[unit] = ipl->ipl_next;
		iplused[unit] -= dlen;
		error = Writemsg(pid, offsetof(struct _io_read_reply, data) + copied, ipl, dlen);
		if (error == -1) {
			ipl->ipl_next = iplt[unit];
			iplt[unit] = ipl;
			iplused[unit] += dlen;
			break;
		}
		KFREES((caddr_t)ipl, dlen);
		ipl_stat[unit].st_size -= error;
	}
	if (!iplt[unit]) {
		iplused[unit] = 0;
		iplh[unit] = &iplt[unit];
		ipll[unit] = NULL;
	}

	return (error == -1) ? -1 : copied;
}


int ipflog_clear(unit)
int unit;
{
	iplog_t *ipl;
	int used;

	MUTEX_ENTER(&ipl_mutex);
	while ((ipl = iplt[unit])) {
		iplt[unit] = ipl->ipl_next;
		KFREES((caddr_t)ipl, ipl->ipl_dsize);
	}
	iplh[unit] = &iplt[unit];
	ipll[unit] = NULL;
	used = iplused[unit];
	iplused[unit] = 0;
	bzero((char *)&iplcrc[unit], FI_CSIZE);
	MUTEX_EXIT(&ipl_mutex);
	return used;
}
#endif /* IPFILTER_LOG */
