/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com for
 * details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */
#ifndef __IPF_RES_H__
#define __IPF_RES_H__

#include <sys/io_msg.h>
#include <sys/stat.h>
#include <sys/sys_msg.h>
#include <stdio.h>
#include <errno.h>

/* This maps the IPL_LOG defininition in ip_fil.h, which is:
 *      #define IPL_LOGIPF      0
 *      #define IPL_LOGNAT      1
 *      #define IPL_LOGSTATE    2
 *      #define IPL_LOGAUTH     3
 *      #define IPL_LOGMAX      3
 */

enum {
	DEV_IPL,
	DEV_IPNAT,
	DEV_IPSTATE,
#ifdef IPFILTER_AUTH
	DEV_IPAUTH,
#endif
	NUM_DEVS
};

struct reader;
struct ocb {
	short int     fd;
	short int     unit;
	mode_t        mode;
	struct reader *rd;
};

struct reader {
	struct ocb      *ocb;
	pid_t           pid;
	struct _io_read msg;
};

struct reader_list_t {
	struct reader *readers;
	int           head;
	int           tail;
	int           maxlen;
};

struct _sysmsg_memcpy {
	caddr_t from;
	int     size;
};

#define _SYSMSG_SUBTYPE_MEMCPY             16

union ipl_msg
{
	short unsigned                          type;
	short unsigned                          status;
	struct {
		struct _sysmsg_hdr                  hdr;
		struct _sysmsg_memcpy               copy;
	}                                       sysmsg;
	struct {
		struct _sysmsg_hdr_reply            hdr_reply;
	}                                       sysmsg_reply;
	struct _io_open                         open;
	struct _io_open_reply                   open_reply;
	struct _io_close                        close;
	struct _io_close_reply                  close_reply;
	struct _io_read                         read;
	struct _io_read_reply                   read_reply;
	struct _io_ioctl                        ioctl;
	struct _io_ioctl_reply                  ioctl_reply;
	struct _io_fstat                        fstat;
	struct _io_fstat_reply                  fstat_reply;
};
extern union ipl_msg iomsg;
extern struct stat	ipl_stat[NUM_DEVS];
extern struct reader_list_t reader_list[NUM_DEVS];

#define REPLY    1
#define NO_REPLY 0

extern int  ipf_res_init(void);
extern void ipf_res_intr(pid_t);
#endif
