/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com for
 * details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <stddef.h>
#include <sys/kernel.h>
#include <sys/io_msg.h>

int ipf_copyin(caddr_t from, caddr_t to, int size, pid_t pid)
{
	int ret = EOK;

	if (pid == 0)
	  memcpy(to, from, size);
	else if (Readmsg(pid, offsetof(struct _io_ioctl, data), to, size) == -1) {
	  ret = errno;
	  syslog(LOG_ERR, "Readmsg (pid = %d, size = %d): %m", pid, size);
	}

	return ret;
}

int ipf_copyout(caddr_t from, caddr_t to, int size, pid_t pid)
{
	int ret = EOK;

	if (pid == 0)
	  memcpy(to, from, size);
	else if (Writemsg(pid, offsetof(struct _io_ioctl, data), from, size) == -1) {
		ret = errno;
		syslog(LOG_ERR, "Writemsg (pid = %d, size = %d): %m", pid, size);
	}

	return ret;
}
