/*
 * Copyright  2000, QNX Software Systems Ltd.  All Rights Reserved
 *
 * This source code has been published by QNX Software Systems Ltd.
 * (QSSL).  However, any use, reproduction, modification, distribution
 * or transfer of this software, or any software which includes or is
 * based upon any of this code, is only permitted under the terms of
 * the QNX Open Community License version 1.0 (see licensing.qnx.com for
 * details) or as otherwise expressly authorized by a written license
 * agreement from QSSL.  For more information, please email licensing@qnx.com.
 */
#include <stdarg.h>
#include <sys/sys_msg.h>
#include <sys/kernel.h>
#include <sys/sendmx.h>
#include <sys/ioctl.h>
#include <sys/qioctl.h>

#include <net/if.h>
#include <netinet/in.h>
#include "netinet/ip_compat.h"
#include "netinet/ip_fil.h"
#include "netinet/ip_auth.h"
#include "netinet/ip_state.h"
#include "netinet/ip_nat.h"

#include "ipf_res.h"

int kmemcpy(fd, buf, pos, n)
int fd;
char *buf;
long pos;
int n;
{
	struct {
		struct _sysmsg_hdr       hdr;
		struct _sysmsg_memcpy    cpy;
	}                            sysmsg;
	struct _sysmsg_hdr_reply     sysmsg_reply;
	struct _mxfer_entry sx[1];
	struct _mxfer_entry rx[2];

	if (!n)
	  return 0;

	sysmsg.hdr.type = _SYSMSG;
	sysmsg.hdr.subtype = _SYSMSG_SUBTYPE_MEMCPY;
	sysmsg.cpy.from = (caddr_t)pos;
	sysmsg.cpy.size = n;

	_setmx(&sx[0], &sysmsg, sizeof(sysmsg));
	_setmx(&rx[0], &sysmsg_reply, sizeof(sysmsg_reply));
	_setmx(&rx[1], buf, n);

	if (Sendfdmx(fd, 1, 2, sx, rx) == -1)
	  return -1;

	if ((errno = sysmsg_reply.status) != EOK)
	  return -1;
	else 
	  return 0;
}

/* this overwrite the orignal ioctl() (by -Dioctl=devctl in Makefile).
 * so that we have to match the ioctl() protocal define in sys/ioctl.h
 */
int devctl(int fd, long cmd, ...)
{
	int sbytes, rbytes;
	void *data;
	va_list v;
	union {
		struct _io_ioctl       s;
		struct _io_ioctl_reply r;
	} msg;
	struct _mxfer_entry sx[2], rx[2];
	int sparts, rparts, off;

	sbytes = rbytes = 0;
	va_start(v, cmd);
	data = va_arg(v, void *);
	va_end(v);

	switch(cmd) {
	  case SIOCATHST:
		rbytes = sizeof(fr_authstat_t);
		break;

	  case SIOCINAFR:
	  case SIOCZRLST:
	  case SIOCINIFR:
	  case SIOCRMIFR:
	  case SIOCADIFR:
		sbytes = sizeof(struct frentry);
		break;

	  case SIOCGETFS:
		rbytes = max(sizeof(ips_stat_t), sizeof(friostat_t));
		break;

	  case SIOCFRZST:
		rbytes = sizeof(friostat_t);
		break;

	  case SIOCGFRST:
		rbytes = sizeof(friostat_t);
		break;

		/* NAT and AFR share the same stupid id !!!
		 case SIOCADNAT:
		 case SIOCRMNAT:
		 */

	  case SIOCRMAFR:
	  case SIOCADAFR:
		sbytes = max(sizeof(ipnat_t), sizeof(struct frentry));
		break;

	  case SIOCGNATS:
		rbytes = sizeof(natstat_t);
		break;

	  case SIOCGNATL:
		sbytes = rbytes = sizeof(natlookup_t);
		break;

	  case SIOCSTPUT:
	  case SIOCSTGET:
		sbytes = rbytes = sizeof(nat_save_t);
		break;

	  case SIOCSTGSZ:
		sbytes = rbytes = sizeof(natget_t);
		break;

#if 0
	  case FIONREAD:
		rbytes = IOCPARM_LEN(cmd);
		save_data = data;
		data = &save_data;
		break;
		/* these should be handled by normal ioctl(). Note the data
		 * pointer here is not a (void **)
		 */
	  case SIOCGETFF:
	  case SIOCSWAPA:
	  case SIOCSTLCK:
	  case SIOCIPFFB:
		rbytes = sizeof(u_int);
		break;

	  case SIOCFRENB:
	  case SIOCSETFF:
	  case SIOCFRSYN:
		sbytes = sizeof(u_int);
		break;

	  case SIOCIPFFL:
		sbytes = rbytes = sizeof(u_int);
		break;
#endif
	  default:
		return ioctl(fd, cmd, data);
	}

	msg.s.type = _IO_IOCTL;
	msg.s.fd   = fd;
	msg.s.zero = cmd;
	msg.s.nbytes = sbytes;

	sparts = rparts = 1;
	off = offsetof(struct _io_ioctl, data);
	_setmx(&sx[0], &msg, off);
	if (sbytes) {
		_setmx(&sx[1], *(void **)data, sbytes);
		sparts++;
	}
//	off = offsetof(struct _io_ioctl_reply, data);
	off = offsetof(struct _io_ioctl, data);
	_setmx(&rx[0], &msg, off);
	if (rbytes) {
		_setmx(&rx[1], *(void **)data, rbytes);
		rparts++;
	}
	
	if (Sendfdmx(fd, sparts, rparts, sx, rx) == -1)
	  return -1;

	if ((errno = msg.r.status) != 0)
	  return -1;
	else
	  return msg.r.ret_val;
}
